package internal

import (
	"github.com/icinga/icinga-go-library/version"
)

// Version contains version and Git commit information.
//
// The placeholders are replaced on `git archive` using the `export-subst` attribute.
var Version = version.Version("1.5.1", "v1.5.1", "7a66f3ad8be0cc8b0edad8028d761eedfc721cab")
