package icingadb_test

import "testing"

func TestStateSync(t *testing.T) {
	// TODO(jb): add tests

	t.Run("Host", func(t *testing.T) { t.Parallel(); t.Skip() })
	t.Run("Service", func(t *testing.T) { t.Parallel(); t.Skip() })
}
